# Reproducibility TOMACS 2023

This documents describes how to reproduce the results discussed in the paper:

"Spatial/Temporal Locality-based Load-sharing in Speculative
Discrete Event Simulation on Multi-core Machine"

Submitted to ACM TOMACS 2023 Special Issue on PADS 2022



## Authors & Contacts

* Federica Montesano <federica.montesano@alumni.uniroma2.eu>
* Romolo Marotta <romolo.marotta@gmail.com>
* Francesco Quaglia <francesco.quaglia@uniroma2.it>

## Requirements

* x86_64 hardware with up to 2 NUMA nodes
* Unix system with gcc toolchain

The hardware configuration used by the authors is:

* OS: Ubuntu 22.04
* CPU: 2x Intel Xeon Silver 4210R CPU @ 2.40GHz
* RAM: 160GB

## Dependencies

* For running tests: ```bash, gcc, cmake, make, libnuma-dev or numactl-devel```
* For processing data: ```bash, Python3, pip3, matplotlib, numpy, scipy, seaborn, scikit-learn```
* For generating RCR report: ```pdflatex```


## Kick the tire instructions

1. Clone the repository: ```git clone --branch TOMACS2023 https://github.com/HPDCS/USE.git USE```
2. ```cd USE/reproducibility/TOMACS2023```
3. Build the project ```./build.sh Release```
4. Run a small simulation ```./use-release/test/test_phold -c 2 -p 1024```

## Structure of the artifact

```
USE/
 |-- ci/                 /* scripts related to software documentation/testing   */
 |-- hijacker/           /* a module for automated production of reverse events */
 |-- script/             /* scripts for hijacker                                */
 |-- models/             /* several simulation models                           */
 |-- src/                /* source code of USE                                  */
 |-- test/               /* unit tests                                          */
 |-- utility-scripts/    /* scripts for analyzing USE source code                       */
 |-- reproducibility/TOMACS2023/    /* contains scripts for RCR activities              */
        |-- figures_original   /* contains figures and tables of the submitted article  */
        |-- scripts_gen        /* contains bash scripts for generating figures   */
        |-- scripts_plot       /* contains python scripts for generating figures */
        |-- scripts_run        /* bash scripts for launching experiments  */
        |-- scripts_table      /* bash scripts for generating tables  */
        |-- autoconf.sh        /* script for generating thread counts for experiments  */
        |-- build.sh           /* script for compiling USE and model binaries  */
        |-- run_exp.sh         /* script for launching a given experiments  */
        |-- process_exp.sh     /* script for   */
        |-- report.tex         /* latex file for generating a RCR report  */
        |-- README.md          /* this readme  */
 |-- AUTHORS.md         /* USE authors          */
 |-- LICENSE            /* USE license          */
 |-- README.md          /* USE usage guidelines */

```
## License

The software is released with the GPL 3 license.

## Article claims

The article has three major claims:

* C1: The spatial/temporal locality approach is effective with small sized pipes
* C2: There is a trade off between improvements achieved by enforcing spatial/temporal locality and rollback incidence
* C3: The overhead on phold simulation is negligible
* C4: The NUMA oriented spatial/temporal locality approach achieves higher performance on balanced models, expecially when they are heavy weight
* C5: The object migrations allows the NUMA oriented spatial/temporal locality approach to achieve higher performance on the long run 
* C6: The object migrations allows the workload skew to be balanced across NUMA nodes 

## Reproducing the results

The paper has 14 Figures and 4 table that can be reproduced.
These can be generated by running 9 experiments.
The mapping between claims, experiments, figures and tables are resumed in the following table.

| Claim | Figures            | Experiment   |
|-------|--------------------|--------------|
| C1    | 5                  | 05               |
| C2    | 6                  | 06               |
| C3    | 7, 8               | 07_and_08        |
| C4    | 9a, 9b, 9c, 12     | 09a, 09b, 09c, 12|      
| C5    | 10a, 10b, 11a, 11b | 10, 11           |
| C6    | 11c, 11d           | 11               |

To setup the environment:
1. ```cd USE/reproducibility/TOMACS2023```
2. ```./build.sh Release```

To run an experiment <exp> and process its results, type the following:

3. ```nohup ./run_exp.sh <exp>```
4. ```nohup ./process_exp.sh <exp>```

To run all experiments and process their results at once, type the following:

3. ```nohup ./run_exp.sh all```
4. ```nohup ./process_exp.sh all```

The expected runtime of each experiment is detailed in the following table:

| Experiment   | Runtime |
|--------------|---------|
| 05           | 1h |
| 06           | 1h 30m |
| 07_and_08    | 3h |
| 09a          | 4h |
| 09b          | 4h |
| 09c          | 4h |
| 10           | 1h |
| 11           | 3h 30m |
| 12           | 2h |
| **Total**    | **26h** |

To generate a report with both original and reproduced figures:

5. ```pdflatex report.tex```

## Notes
All scripts have been tested by running them from the following path:

  ```USE/reproducibility/TOMACS2023```
